{% extends '!layout.html' %}

{% block sidebarlogo %}
  {% if pagename == 'index' %}
    <iframe
      src="http://ghbtns.com/github-btn.html?user=emcconville&repo=wand{{- '' -}}
           &type=watch&count=true&size=large"
      allowtransparency="true" frameborder="0" scrolling="0"
      width="170px" height="30px">
      {{- '' -}}
    </iframe>
    <hr class="github-line" />
  {% else %}
    {{ super() }}
  {% endif %}
{% endblock %}
