
@import url(http://fonts.googleapis.com/css?family=Sunshiney);
@import url(flasky.css);

.indexwrapper .sphinxsidebar .github-line { visibility: hidden; }

.indexwrapper div.body #wand.section h1 {
  background-image: url(wand.png);
  background-size: 75px 75px;
  background-repeat: no-repeat;
  padding-left: 75px;
  padding-top: 10px;
  height: 75px;
}

.logo img.logo { width: 120px; height: 120px; padding-right: 30px; }

div.body h1, div.body h2, div.body h3, div.body h4, div.body h5, div.body h6
  { font-family: 'Sunshiney', 'Garamond', 'Georgia', serif; }

.literal-block .subscript { vertical-align: sub; }
.literal-block .supscript { vertical-align: super; }